/*
** conan controller 0.1 atmega128 16 MHz
** S. James Remington, sjames_remington at yahoo dot com
*/

/* CPU frequency */
#define F_CPU 16000000UL
/* UART baud rate and input buffer size */
#define UART_BAUD  9600
#define RX_BUFSIZE 60

#include <ctype.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <avr/io.h>
#include <avr/pgmspace.h>

#include <avr/interrupt.h>
#include <util/delay.h>

//GLOBALS

unsigned int ticks,seconds;				//~1ms ticks, total run time
long int enc0_pos,enc1_pos,enc2_pos,enc3_pos;  //encoder position counters
int	enc0_per,enc1_per,enc2_per,enc3_per,left_spd,right_spd;
unsigned char enc0_dir,enc1_dir,enc2_dir,enc3_dir; //current encoder direction of rotation

extern int uart_putchar(char c, FILE *stream);
extern int uart_getchar(FILE *stream);

FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

/*

  Timer0 overflow interrupt handler. Counts system clock/64/256 -- provides global ticks at 976.56 Hz

*/

ISR(TIMER0_OVF_vect)
{
   static unsigned int pcount = 977;  //ticks per second (roughly)
	ticks++;
   if (pcount-- == 0) {  //decrement tick counter and update runtime seconds
	pcount=977;
    seconds++;
	}
}

/*

 INT5 (PORTE.5) interrupt service routine. Interrupt occurs on rising edge of INT5

 -Assumes encoder0 PORTE.5=A channel, PORTD.6=B channel
 -Increment or decrement encoder 0 position accordingly
 -Keep running period (time between interrupts) low pass filter New = (0.75*old) + (0.25)*current
 
 */

ISR (INT5_vect) {
	static unsigned int last_tick=0;
	int tmp;

	tmp = (signed int) (ticks-last_tick);				//determine period, time since last stripe
	if(tmp>0) enc0_per=((enc0_per*3) + tmp) >> 2;		//low pass filter, skip if tick timer overflow
	last_tick=ticks;									//save time stamp
	enc0_dir = PIND & _BV(6);							//determine and update direction of rotation
	if(enc0_dir==0) enc0_pos++; else enc0_pos--;		//update encoder position
}

/*

 Timer1 overflow interrupt handler. Turns on the pulse to the left motors if requested speed <> 0

*/

ISR (TIMER1_OVF_vect)
{
	if(left_spd != 0)	PORTB |= (_BV(4)|_BV(5));  //turn on PWM to left side motors
}
/*

 Timer 1 OUTPUT COMPARE A Interrupt
 This interrupt occurs on compare match of TCNT1 with OCR1A.
 set motor 0 PWM to off

*/

ISR (TIMER1_COMPA_vect)
{
	PORTB &= ~_BV(4);  //PB4 = motor0 PWM = 0
}

/*

 Timer 1 OUTPUT COMPARE B Interrupt
 This interrupt occurs on compare match of TCNT1 with OCR1B.
 set motor 3 PWM to off

*/

ISR (TIMER1_COMPB_vect)
{
	PORTB &= ~_BV(5); 	//PB5 = motor3 PWM = 0
}

/*

 Timer3 overflow interrupt handler. Turns on the PWM pulse to the right motors if speed <> 0

*/

ISR (TIMER3_OVF_vect)
{
	if(right_spd != 0)	PORTE |= (_BV(3)|_BV(4));  //turn on right side motors
}
/*

 Timer 3 OUTPUT COMPARE A Interrupt
 This interrupt occurs on compare match of TCNT3 with OCR3A.
 set motor 1 PWM to off

*/

ISR (TIMER3_COMPA_vect)
{
	PORTE &= ~_BV(3);  //PE3 = motor1 PWM = 0
}

/*

 Timer 3 OUTPUT COMPARE B Interrupt
 This interrupt occurs on compare match of TCNT3 with OCR3B.
 set motor 2 to off

*/

ISR (TIMER3_COMPB_vect)
{
	PORTE &= ~_BV(4); 	//PE4 = motor2 PWM = 0
}

/*

  Initialize Timers 0,1,3

*/

static void timer_init(void)
{

// set up Timer0 to provide ~1 ms ticks

	TCCR0 = _BV(CS02);  //clock/64
	TIMSK |= _BV(TOIE0); // enable int on overflow

// set up Timer1 to provide fast PWM mode 5 8-bit output for motors

    TCCR1A = ( 1 << WGM10 );
    TCCR1B = ( 3 << CS10  ) | ( 1 << WGM12 );  //clock/64 /256 => 976.56 Hz
	OCR1A = 0;
	OCR1B = 0;

// set up Timer3 to provide fast PWM mode 5 8-bit output for motors

    TCCR3A = ( 1 << WGM30 );
    TCCR3B = ( 3 << CS30  ) | ( 1 << WGM32 );  //clock/64 /256 => 976.56 Hz
	OCR3A = 0;
	OCR3B = 0;

// enable interrupts on timer 1 and 3 overflow and output compares A and B

	TIMSK  |= (_BV(TOIE1) | _BV(OCIE1A) | _BV(OCIE1B));
	ETIMSK |= (_BV(TOIE3) | _BV(OCIE3A) | _BV(OCIE3B));
}

/*
 include the UART1 routines 
*/

#include "uart.c"

/*

  post requested speeds for left and right motors as globals -- for later PID algorithm

*/

void set_speed(int left, int right) {

		  if (left<0) { 
		  	PORTD &=~(_BV(4)|_BV(5));	//set appropriate motor 0 and 3 direction
			left=-left; 				// if negative, make speed positive
			}
		  else PORTD |= _BV(4)|_BV(5); 
		  if (left>255) left=255;		// max allowed = 255
		left_spd=left;
		  OCR1A=OCR1B=left;				// set output compare registers
		  
		  if (right<0) { 
		  	PORTB |= (_BV(2)|_BV(3));  	//set appropriate motor 1 and 2 direction
			right=-right; 				//make speed positive if neg
			}
		  else PORTB &= ~(_BV(2)|_BV(3)); 
		  if(right>255) right=255;		// max allowed = 255
		right_spd=right;
		  OCR3A=OCR3B=right;
}

/*
	MAIN PROGRAM
*/

int main(void) {

	char buf[10];
	int arg,err,i;
	
// initialize ports ASAP


	DDRB |= (_BV(0)|_BV(2)|_BV(3)|_BV(4)|_BV(5));		//PB 0=LED output; 
							//PB4,5 set pins for motors 0 and 1 to output
							//PB2,3 set pins for motor dir 1 and dir 2

  	PORTB = 1;  //led on, set motors 0 and 3 to off, set motor dir 1 and 2 to 0

	DDRD |= _BV(4)|_BV(5);						//PD4,5 set port pins for motor dir 0 and 3 to output
	PORTD &= ~(_BV(4)|_BV(5));					//set to zero
	
	DDRE |= _BV(3)|_BV(4);						//PE3,4 set port pins for motors 1 and 2 to output
	PORTE &= ~(_BV(3)|_BV(4));					//PE3,4 set motors 1 and 2 off

	timer_init();

// enable external interrupts

	EICRB |= (3<<ISC50); 				//enable encoder0_A interrupt on rising edge (PORTE.5)
	EIMSK |=_BV(5);						//enable encoder 0 (INT5) interrupt

// enable uart

	uart_init();
	stdout = stdin = &uart_str;			// associate stream with stdout, stdin

	for(i=0; i<10; i++) _delay_ms(10);		//delay 100 ms for uart to wake up

	printf_P(PSTR("\nCONAN, the Warrior Robot 0.2\nValid commands: hmpt\n"));

	seconds=0;
	
	sei(); //enable interrupts

//  ***** control code starts here


	for (;;) {

	printf("%d>",seconds);  //prompt for command, print running seconds counter for debug traceback
	fgets(buf, sizeof buf - 1, stdin);

	err=0;						//initialize error flag
	
	switch (tolower(buf[0]))
		{
		default:
		printf("?: %s\n", buf);
		err=1;
		break;

//m, set motor speed. currently one argument for both left and right

		case 'm':  //motor speed, range 0 to +/-255 (0 off)
		if (sscanf(buf, "%*s %d", &arg) > 0) {  //if argument on hand,
			printf("%u: PW= %d\n", seconds, arg);
			set_speed(arg,arg);
			}
		else err=2;  //no argument
		break;

//p, print encoder position
		
		case 'p':
		printf("Enc0 pos = %ld\n",enc0_pos);
		break;

//t, print encoder period

		case 't':
		printf("Enc0 per = %d\n",enc0_per);
		break;

//h, HALT
		
		case 'h':
		printf("%u: Halt\n",seconds);

// ramp speed down. Regardless of direction of rotation, OCRnx is positive. 

		arg=OCR1A;
		if (arg<OCR3A) arg=OCR3A;	//choose approximate max of right or left pwm value

		for(i=arg; i>-1; i -= 5) {OCR1A=OCR1B=i; _delay_ms(5);}	//ramp down

		OCR1A=OCR1B=0;
		OCR3A=OCR3B=0;

		break;
		}
	
// bad argument in last command

	if(err==2) printf("Arg? %s \n",buf);

	}

	printf("%u: All done...",seconds);		//currently, never reaches this point

	return 0;
}
