/* 12/21/2006 robot_control5
** serial command-driven servo controller, baby Orangutan
** This version first parses then buffers commands for later execution. Interrupt vectors updated.
** PD speed control and stall timer added 12/21
** I2C routines added for compass 12/19/06
** PC0 = forward steering servo pulse
** PC1 = reverse steering servo pulse
** PC2 = motor control pulse
** PC4,5 = SDA, SCL for I2C bus
** PD2, 3 = int0,int1 interrupts. int0 = wheel
*/

/* CPU frequency */
#define F_CPU 20000000UL
/* UART baud rate */
#define UART_BAUD  9600

// timer 1 setting for 1.5 ms neutral pulses in units of 0.01 ms
// = (150 input/100 units/ms)*6250 (incs/sec)/(20 ms/timer cycle)=468  -- prescaler 64 @ 20 MHz
#define T1_OCR_N (150*625UL)/(200UL)
// timer 2 setting for 1.5 ms neutral pulses in units of 0.05 ms (20 to 40)
// = ( 30 input/20 units/ms)*255 (incs/sec)/(13 ms/timer cycle)=29   -- prescaler 1024 @ 20 MHz
#define T2_OCR_N (15*255UL)/(10UL*13UL)
// number of 1/50 second clock ticks before wheel stall timeout
#define wheel_stall_timeout 25
//convert wheel tick period to rotational speed. Select so that max=10
#define SPDSCALE 30000	

#include <ctype.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <avr/io.h>
#include <avr/pgmspace.h>

#include <avr/interrupt.h>
#include <util/delay.h>

#include "robot_control.h"
#include "uart.c"
#include "twimaster.c"

// external io routines become stdio

extern int uart_putchar(char c, FILE *stream);
extern int uart_getchar(FILE *stream);
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

//GLOBALS

unsigned int seconds,ticks,halt_time,wheel_timeout,err;			//total run timers and stall timer
unsigned int PW_F,PW_R,PW_M;						//pulse on times for servos

//Navigation
signed int dir,speed,req_speed;		//direction of travel, current and requested speeds
signed int bearing,wheel_pos,req_wheel_pos;	//compass bearing 0-359, position: 2 inches/increment = +/- 5450 feet max


//Timer0 overflow interrupt handler. Counts system clock/64/256 -- ticks at 1220.7 Hz
ISR(TIMER0_OVF_vect)
{
	ticks++;
}

// Timer1 overflow interrupt handler. Turns on the pulse to the F and R servos
// updates out_F and out_R if changed, increments runtime seconds counter.

ISR(TIMER1_OVF_vect)
{
	static char secs_frac=0;

	PORTC |= (_BV(0)|_BV(1));				//F=PC0, R=PC1
	OCR1A = PW_F;
	OCR1B = PW_R;

	if (secs_frac++ == 50) {				//keep time and halt when "there"
	secs_frac=0;
    seconds++;
//	printf("p %d\n",speed);
	if (seconds == halt_time) exec('h',1,0);
	}
/*
	if(req_speed > 0) {						//check for stall condition, if supposed to be moving
	wheel_timeout--;
		if(wheel_timeout==0) {					//stall detected. Increase power
		wheel_timeout=wheel_stall_timeout;
		req_speed++;						//increment req_speed and set new PWM value
		PW_M=T2_convert(req_speed);
		}
	}	
*/
}
// OUTPUT COMPARE 1A Interrupt
// This interrupt occurs on compare match of TCNT1 with OCR1A.
// set forward steering servo low

ISR(TIMER1_COMPA_vect)
{
	PORTC &= ~_BV(0);  //PC0=0
}

// OUTPUT COMPARE 1B Interrupt
// This interrupt occurs on compare match of TCNT1 with OCR1B.
// set reverse steering servo low

ISR(TIMER1_COMPB_vect)
{
	PORTC &= ~_BV(1); 	//PC1=0
}

// Timer2 overflow handler

ISR(TIMER2_OVF_vect)
{
	PORTC |= (_BV(2));  //M=PC2=1
	OCR2A = PW_M;
}

//Timer2 Output Compare A interrupt handler
//Motor control servo line set low

ISR(TIMER2_COMPA_vect)
{
	PORTC &= ~_BV(2);  //M=PC2=0
}

/*
** INTERRUPT0 ISR
** This routine is called on each pulse of the wheel encoder
** now checks for endpoint reached
*/

ISR(INT0_vect)
{
	static unsigned int tmr_prev=0;
	static unsigned int enc_per=0;
	volatile unsigned int tmr;

	tmr=ticks;
	wheel_timeout=wheel_stall_timeout;  //wheel is rotating so reset stall timer

	if(req_speed>0) wheel_pos++;
	else if (req_speed<0) wheel_pos--;

	if(req_wheel_pos == wheel_pos) {
	exec('h',1,0);  //halt if at requested position
	req_wheel_pos=-32000;  //so we don't stop there again
	}

// calculate weighted moving average velocity
	if (tmr>tmr_prev) {  //skip period update if tick counter overflowed last time around
	enc_per = (enc_per*3 + (tmr-tmr_prev))>>2;
	speed=enc_per;
	}
	tmr_prev=tmr;
}


void hardware_init(void)
{
	TCCR0A = 0; //tmr0 mode 0 - normal timer operation
	TCCR0B = _BV(CS01)|_BV(CS00);  //clock/64
	TIMSK0 = _BV(TOIE0); // enable int on overflow

	TCCR1A = _BV (WGM11); //tmr1 is PWM with TOP set by ICR1 in fast PWM mode 14
	TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS11) | _BV (CS10); //tmr1 running on fosc/64 = 312,500 Hz clock
	ICR1 = 6250; //set PWM frequency to 50Hz = 20 ms period
	PW_F = T1_OCR_N;//set duty cycle such that 1.5 ms pulses are generated; 1.5ms = 468
	OCR1A = PW_F;
	PW_R = T1_OCR_N;
	OCR1B = PW_R;
	TIMSK1 |= (_BV (TOIE1) | _BV(OCIE1A) | _BV(OCIE1B)); //enable ints on overflow & output compares

	TCCR2A = _BV (WGM21)|_BV(WGM20); //tmr2 is PWM with TOP=255 in PWM mode 17
	TCCR2B = _BV(CS22) | _BV(CS21) | _BV (CS20);  //fosc/1024
	PW_M = T2_OCR_N; //MAX=PWM frequency 76.6 Hz = 13.06 ms period
	OCR2A = PW_M; //set duty cycle such that 1.5 ms pulses are generated; 1.5ms = 29, 2ms = 39
	OCR2B = 0;  //not used
	TIMSK2 |= (_BV (TOIE2) | _BV(OCIE2A)); //enable ints on overflow and output compare A

//	PORT INITIALIZATION

	DDRD &= ~_BV(2);			//INT0=PD.2=input
	
	DDRD |= (_BV(5)|_BV(6));	//Motor control PD outputs
	PORTD &= ~(_BV(5)|_BV(6));	//Set both low
	DDRB |= (_BV(1)|_BV(2));	//Motor control PB outputs
	PORTB &= ~(_BV(1)|_BV(2));	//Set both low so that H-bridges do not draw base current
	
	DDRC=7;			//PC0,1,2 = output

	//enable interrupt 0 for rising and falling edge trigger (10 counts/rev)
	EICRA |= _BV(ISC00);
	EIMSK |= _BV(INT0);

}

/* COMMAND EXECUTION
** exec -- execute a parsed command.
** returns 0 if no errors 
** err=1 unknown command
** err=2 invalid argument
*/

int exec(char c, unsigned char numarg, int arg)
{
	volatile int t0;

	err=0;						//initialize error flag
	
	switch (c)
	{
		default:
		printf("?: %c\n", c);
		err=1;
		break;
		
//f: set forward steering servo pulse width (100-200)
		case 'f': 
		if (numarg > 0) {
			PW_F=T1_convert(arg);
			printf("%u: %c %d\n", seconds,c, arg);
			}
		else err=2;
		break;
		
//r: set rear steering servo pulse width (100-200)
		case 'r':
		if (numarg > 0)	    {
			PW_R=T1_convert(arg);
			printf("%u: %c %d\n", seconds,c, arg);
			}
		else err=2;
		break;
		
//p: set motor power, range -10 to 10. Also separates and makes global dir and req_speed.
		case 'p':
		if (numarg > 0)	    {
			dir=0; req_speed=0;
			if(arg>0) {dir= 1; req_speed= arg;}
			if(arg<0) {dir=-1; req_speed=-arg;}
			PW_M=T2_convert(arg);
			printf("%u: %c %d\n", seconds,c, arg);
			}
		else err=2;
		break;
		
//s: set steering angle, range -30 to 30		
		case 's':
		if (numarg > 0)	    {
			if (arg< -30) arg=-30;
			if (arg> 30) arg=30;
			PW_F=T1_convert(150+arg);
			PW_R=T1_convert(150-arg);
			printf("%u: %c %d\n", seconds,c, arg);
			}
		else err=2;
		break;
		
//z: reset running seconds timer. If there is an argument, set wheel_pos to its value.		
		case 'z':
		printf("%d: T reset", seconds);
		seconds=0;
		halt_time=0;
		if (numarg>0) {
			wheel_pos=arg;
			printf(", odometer set: %d\n",arg);
			}
		else puts("");  //newline
		break;
		
//w: read or set requested wheel position (odometry). Does not check direction of travel!
		case 'w':
		if(numarg>0) {  //if argument given, set requested. Else print current position
			req_wheel_pos=wheel_pos+arg;
			printf("%d: Destination %d\n",seconds,req_wheel_pos);
			}
		else printf("pos= %d\n",wheel_pos);
		break;
		
//c: continue or pause for nnn seconds
		case 'c':  //continue or pause, fall through
		
		if (numarg > 0)	    {
			if(arg<0) arg=-arg;
			printf("%u: Continue %d", t0=seconds, arg);
			while( seconds < (t0+arg));
			puts("");  //just a <crlf>
			}
		else err=2;
		break;
		
//h: halt, reset steering servos to neutral and motor to off
		case 'h':
		printf("\n%u: Halt at position %d\n",seconds,wheel_pos);
		PW_M = T2_convert(0);
		PW_F = PW_R = T1_convert(150);
		break;

//b: read compass bearing
		case 'b':
		{
		volatile unsigned char highbyte,lowbyte;
			i2c_start_wait(0xC0+I2C_WRITE);
			i2c_write(2);  //two-byte bearing = register 2&3
			i2c_rep_start(0xC0+I2C_READ); //set read mode
			highbyte=i2c_readAck();	//read high byte
			lowbyte=i2c_readNak();
			i2c_stop();
			bearing = (unsigned int) (highbyte<<8) + (unsigned int) lowbyte;
			printf("%u: bearing= %u.%u\n",seconds,bearing/10,bearing % 10);
		break;
		}
		
//t: set timeout to halt (seconds)
		case 't':
		if (numarg>0) {
			if(arg<0) arg=-arg;
			halt_time=seconds+arg;
			printf("%u: timeout halt in %u seconds\n",seconds,halt_time);
			}
		else err=2;
		break;
		
	}  //end switch
	
	if(err==2) printf("%c Arg?\n",c);
	return err;
}

int main(void) {

	char c1,c2,buf[30],ctable[10];
	unsigned char narg,numarg[10],ncmd;
	int i,arg,arg2,arg3,argtable[10];
	unsigned int w;

	uart_init();
	hardware_init();
	i2c_init();  // initialize I2C interface

  
	stdout = stdin = &uart_str;
//	_delay_ms(250);
//	_delay_ms(250);
	for (w=1;w<65535;w++) i=w;
	
	printf_P(PSTR("\nRobot controller 1.4!\nValid commands: hmsp fr cbzwt []\n"));
	err=0;
	seconds=0;
	halt_time=0;
	wheel_pos=0;
	req_speed=0;
	dir=0;					//current direction of movement
	req_wheel_pos=-32000;  //an unlikely place to be.
	sei(); //enable interrupts


	for (;;) {

	printf("%d>",seconds);

//	if (fgets(buf, sizeof buf - 1, stdin) == NULL) break;  //note: won't break, <cr> is included for terminal input

	fgets(buf, sizeof buf - 1, stdin);  // fetch and parse a command
	c1=tolower(buf[0]);
	arg=0; //default argument
	narg=sscanf(buf, "%*s %d %d %d", &arg,&arg2,&arg3);
	err=0;
	switch (c1) {

//[:  fetch, parse and store additional commands until "]" command encountered
		case '[': 
			ncmd=0;
			for(;;) {
			printf("%d.",ncmd);
			fgets(buf, sizeof buf - 1, stdin); //get and save a command
			if( (c2=tolower(buf[0])) == ']') break;
			else {
			arg=0; //default argument
			narg=sscanf(buf, "%*s %d", &arg);	//get arguments, if any

				if (ncmd<10) {  					//do not store more than 10 commands
					ctable[ncmd]=c2;
					numarg[ncmd]=narg;
					argtable[ncmd]=arg;
					ncmd++;
					}
				}
			}
		break;

//g:  execute stored commands, as many times as specified by command argument (but at least once)
		case 'g':
				do {
					if(ncmd>0) {  //execute stored commands
					for(i=0; i<ncmd; i++) {
						printf("[%d] ",i);
						if ( (err=exec(ctable[i],numarg[i],argtable[i]))>0) break;
						}
					arg--;
					}
				}
				while (arg>0);
		break;  //exit execute loop

		
//m: MoveRel. Takes 3 arguments: steering angle (-30 to 30), speed (-10 to 10), wheel ticks(>0) to destination
		case 'm':
		if (narg == 3) {
			printf("%u: Move: %d %d %d\n",seconds,arg,arg2,arg3);
			exec('s',1,arg);
			exec('p',1,arg2);
			if(dir<0) arg3=-arg3;
			exec('w',1,arg3);	//destination, taking travel direction into account
			break;
			}
		else {
			err=2;
			break;
			}
		
		case 'x':
		printf("speed %d",speed);
		break;
		
// default, call exec with parsed command, first argument 
		default: 
		err=exec(c1,narg,arg);
		break;

		}  //end switch

		if (err>0) 
		{
		printf("Error #%u ... halt\n",err);
		PW_M = T2_convert(0);
		PW_F = PW_R = T1_convert(150);
		}

	}  // end for(;;)
//	printf("%u: All done, thank you...",seconds);
	return 0;
}

//T1_convert: convert .01 millisecond pulse width units (150=1.5 ms) to timer ticks
//allowed range is 100-200

unsigned int T1_convert(unsigned int input)

{
	volatile unsigned int temp;
// pw*(6250ticks)/(20ms*100)  exceeds unsigned integer range, so compute 62.5/20 
	temp=input;
	if (temp<100) temp=100;
	if (temp>200) temp=200;
	return (temp*62+temp/2)/20;
}

//T2_convert: convert .05 millisecond pulse width units (30=1.5 ms) to compare register ticks
//allowed range is -10 to 10, adds 30 to put into range 20-40

unsigned int T2_convert(unsigned int input)
{
	volatile unsigned int temp;
// pw*(255ticks)/(13ms*20) 
// All this just ends up subtracting 1 from input, if in range 20-40
	temp=input+30;
	if (temp<20) temp=20;
	if (temp>40) temp=40;
	return (temp*255)/260;
}

// returns -1,0,1 sign of integer value (ala BASIC)
int sign(int value)
{
	volatile int s;
	s=0;
	if(value>0) s=1;
	if(value<0) s=-1;
	return s;
}
